//----------------------------------------------------------------------------
//  EDGE OpenGL Rendering (BSP Traversal)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -AJA- 2000/01/03: Wrote this file (by cleaning up r2_bsp.c).
//
// TODO HERE:
//   + optimise first subsector: ignore floors out of view.
//   + split up: rgl_seg.c and rgl_mobj.c.
//   + handle scaling better.
//   - implement halos.
//

// this conditional applies to the whole file
#ifdef USE_GL

#include "i_defs.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "e_search.h"
#include "m_bbox.h"
#include "m_random.h"
#include "p_local.h"
#include "p_mobj.h"
#include "r_defs.h"
#include "r_main.h"
#include "r_plane.h"
#include "r_sky.h"
#include "r_state.h"
#include "r_things.h"
#include "r2_defs.h"
#include "rgl_defs.h"
#include "v_colour.h"
#include "v_res.h"
#include "w_textur.h"
#include "w_wad.h"
#include "v_ctx.h"
#include "z_zone.h"


#define DEBUG  0

#define HALOS    0


// colour of the player's weapon
int rgl_weapon_r[4];
int rgl_weapon_g[4];
int rgl_weapon_b[4];

//!!!!!
extern int glpar_invuln;
extern int glpar_dither;

//!!!!!! TMP HACK
#define MAX_TEX 10000
extern stored_gl_tex_t conv_sprites[MAX_TEX];


// common stuff

extern sector_t *frontsector;
extern sector_t *backsector;

static subsector_t *cur_sub;
static seg_t *cur_seg;

static boolean_t cur_upper_sky;
static boolean_t cur_lower_sky;
static boolean_t cur_upper_invis;
static boolean_t cur_lower_invis;

static boolean_t solid_mode;

//!!!!! HACK FIXME
extern boolean_t ren_allbright;
extern float_t ren_red_mul;
extern float_t ren_grn_mul;
extern float_t ren_blu_mul;

static subsector_t *drawsubs_head;
static subsector_t *drawsubs_tail;

int rgl_currtimeval;  // hack..


// The minimum distance between player and a visible sprite.
#define MINZ        (4.0)


//
// LIGHTING RULES:
//
//   1. N/S/E/W changes and infrared are applied before darkening.
//   2. extralight and dlights are applied after darkening.
//   3. for gamma, we rely on the gamma function g() having the
//      property: g(X * Y) = g(X) * g(Y).
//

static INLINE int RGL_Light(int nominal)
{
  if (ren_allbright)
    return 255;
  
  if (effect_infrared)
    nominal += (int)(effect_strength * 255);

  DEV_ASSERT2(currentmap);
  if (currentmap->lighting <= LMODEL_Doomish)
  {
    nominal = rgl_light_map[MIN(255, nominal)];
  }
  
  nominal += extralight * 16;
  
  return GAMMA_CONV(MIN(255, nominal));
}

#if 0  // -AJA- not used anymore, but small chance it could be.
static INLINE float_t RGL_Shader(int nominal, float_t z)
{
  lighttable_t *table;

  if (ren_allbright)
    return 1.0;
  
  nominal >>= LIGHTSEGSHIFT;

  table = zlight[MAX(0, MIN(LIGHTLEVELS-1, nominal))];

  nominal = (int) (z / (1 << (LIGHTZSHIFT - FRACBITS)));
  nominal = MAX(0, MIN(MAXLIGHTZ-1, nominal));

  return MIN(255, table[nominal] + extralight * 16) / 255.0;
}
#endif

#define LT_RED(light)  (MIN(255,light) * ren_red_mul / 255.0)
#define LT_GRN(light)  (MIN(255,light) * ren_grn_mul / 255.0)
#define LT_BLU(light)  (MIN(255,light) * ren_blu_mul / 255.0)


//
// RGL_DrawWall
//
static void RGL_DrawWall(drawfloor_t *dfloor, float_t top,
    float_t bottom, float_t tex_top_h, sidepart_t *part,
    boolean_t masked, boolean_t opaque, float_t x_offset)
{
  float_t x1 = cur_seg->v1->x;
  float_t y1 = cur_seg->v1->y;
  float_t x2 = cur_seg->v2->x;
  float_t y2 = cur_seg->v2->y;

  float_t tex_x1 = 0, tex_y1 = 0;
  float_t tex_x2 = 0, tex_y2 = 0, tex_dy = 0;

  int w, h, i;
  float_t t_right, t_bottom;
  float_t trans = part->translucency;

  boolean_t blended;

  GLuint tex_id;
  const cached_image_t *cim;

  local_gl_vert_t *vert, *orig;

  int L_r[4], L_g[4], L_b[4];

  int lit_Nom = dfloor->reg->p->lightlevel;
  const colourmap_t *colmap = dfloor->reg->p->colourmap;
  float_t c_r, c_g, c_b;

  // ignore non-solid walls in solid mode (& vice versa)
  blended = (part->translucency <= 0.99) ? true : false;
  if ((blended || masked) == solid_mode)
    return;
  
  DEV_ASSERT2(currentmap);
  
  // do the N/S/W/E bizzo...
  if (currentmap->lighting == LMODEL_Doom)
  {
    if (cur_seg->v1->y == cur_seg->v2->y)
      lit_Nom -= 8;
    else if (cur_seg->v1->x == cur_seg->v2->x)
      lit_Nom += 8;
  }

  V_GetColmapRGB(colmap, &c_r, &c_g, &c_b, false);

  lit_Nom = RGL_Light(lit_Nom);

  for (i=0; i < 4; i++)
  {
    L_r[i] = lit_Nom * c_r;
    L_g[i] = lit_Nom * c_g;
    L_b[i] = lit_Nom * c_b;
  }

  DEV_ASSERT2(part->image);
  cim = W_ImageCache(part->image, IMG_OGL, 0, true);
  tex_id = W_ImageGetOGL(cim);

  // Note: normally this would be wrong, since we're using the GL
  // texture ID later on (after W_ImageDone).  The W_LockImagesOGL
  // call saves us though.
  W_ImageDone(cim);

  w = part->image->actual_w;
  h = part->image->actual_h;
  t_right  = w / (float_t)part->image->total_w;
  t_bottom = h / (float_t)part->image->total_h;

  tex_x1 = cur_seg->offset + x_offset;
  tex_x2 = tex_x1 + cur_seg->length;
  
  tex_y1 = tex_top_h - top;
  tex_y2 = tex_top_h - bottom;
  tex_dy = part->skew * cur_seg->length;

  tex_x1 /= (float_t) w * part->scale;
  tex_x2 /= (float_t) w * part->scale;

  tex_y1 /= (float_t) h * part->scale;
  tex_y2 /= (float_t) h * part->scale;
  tex_dy /= (float_t) h * part->scale;

#if (DEBUG >= 3) 
    I_WriteDebug( "WALL (%d,%d,%d) -> (%d,%d,%d)\n", 
      (int) x1, (int) y1, (int) top, (int) x2, (int) y2, (int) bottom);
#endif

  //
  //  Dynamic Lighting Stuff
  //

  if (use_dlights)
  {
    divline_t div;
    drawthing_t *dl;
    float_t wx[4], wy[4], wz[4];

    wx[0] = x1;  wy[0] = y1;  wz[0] = bottom;
    wx[1] = x1;  wy[1] = y1;  wz[1] = top;
    wx[2] = x2;  wy[2] = y2;  wz[2] = top;
    wx[3] = x2;  wy[3] = y2;  wz[3] = bottom;

    div.x  = cur_seg->v1->x;
    div.y  = cur_seg->v1->y;
    div.dx = cur_seg->v2->x - div.x;
    div.dy = cur_seg->v2->y - div.y;

    for (dl=dfloor->dlights; dl; dl=dl->next)
    {
      // light behind seg ?    
      if (P_PointOnDivlineSide(dl->mo->x, dl->mo->y, &div) != 0)
        continue;
       
      R2_AddColourDLights(4, L_r, L_g, L_b, wx, wy, wz, dl->mo);
    }
  }

  vert = orig = RGL_BeginUnit(GL_QUADS, 4, tex_id, masked, blended);

  SET_COLOR(LT_RED(L_r[0]), LT_GRN(L_g[0]), LT_BLU(L_b[0]), trans);
  SET_TEXCOORD(tex_x1, 1.0 - t_bottom * tex_y2);
  SET_NORMAL(y2 - y1, x1 - x2, 0.0);
  SET_EDGE_FLAG(GL_TRUE);
  SET_VERTEX(x1, y1, bottom);
  vert++;

  SET_COLOR(LT_RED(L_r[1]), LT_GRN(L_g[1]), LT_BLU(L_b[1]), trans);
  SET_TEXCOORD(tex_x1, 1.0 - t_bottom * tex_y1);
  SET_NORMAL(y2 - y1, x1 - x2, 0.0);
  SET_EDGE_FLAG(GL_TRUE);
  SET_VERTEX(x1, y1, top);
  vert++;

  SET_COLOR(LT_RED(L_r[2]), LT_GRN(L_g[2]), LT_BLU(L_b[2]), trans);
  SET_TEXCOORD(tex_x2, 1.0 - t_bottom * (tex_y1 + tex_dy));
  SET_NORMAL(y2 - y1, x1 - x2, 0.0);
  SET_EDGE_FLAG(GL_TRUE);
  SET_VERTEX(x2, y2, top);
  vert++;

  SET_COLOR(LT_RED(L_r[3]), LT_GRN(L_g[3]), LT_BLU(L_b[3]), trans);
  SET_TEXCOORD(tex_x2, 1.0 - t_bottom * (tex_y2 + tex_dy));
  SET_NORMAL(y2 - y1, x1 - x2, 0.0);
  SET_EDGE_FLAG(GL_TRUE);
  SET_VERTEX(x2, y2, bottom);
  vert++;

  RGL_EndUnit(vert - orig);
}


//
// RGL_BuildWallSection
//
static INLINE void RGL_BuildWallSection(drawfloor_t *dfloor,
    vert_region_t *reg, float_t top, float_t bottom,
    int line_flags, float_t f0, float_t c0, float_t *f1, float_t *c1)
{ 
  float_t f, c, tex_top_h, x_offset;
  sidepart_t *part;
  boolean_t opaque = false;

  // not visible ?
  if (top <= bottom || bottom >= (*c1) || top <= (*f1))
    return;
  
  if (! reg)
  {
    if (cur_lower_sky)
      return;

    // lower part
    part = &cur_seg->sidedef->bottom;
    x_offset = part->x_offset;
    opaque = true;
    
    tex_top_h = (line_flags & ML_DONTPEGBOTTOM) ?
        c0 : backsector->floor.h;
    tex_top_h += part->y_offset;    
  }
  else if (!reg->sec_next || !reg->extrafloor)
  {
    if (cur_upper_sky)
      return;

    // upper part
    part = &cur_seg->sidedef->top;
    x_offset = part->x_offset;
    opaque = true;

    if (! part->image)
      return;
    tex_top_h = (line_flags & ML_DONTPEGTOP) ?
        c0 : backsector->ceil.h + part->image->actual_h;
    tex_top_h += part->y_offset;    
  }
  else
  {
    DEV_ASSERT2(reg->extraline);
    DEV_ASSERT2(reg->extraline->side[0]);

    // -AJA- Don't bother drawing extrafloor sides if the front/back
    //       sectors have the same tag (and thus the same
    //       extrafloors).
    if (backsector && frontsector->tag == backsector->tag)
      return;

    part = &reg->extraline->side[0]->middle;
    x_offset  = cur_seg->sidedef->middle.x_offset;

    tex_top_h = (reg->extraline->flags & ML_DONTPEGBOTTOM) ?
        reg->c_h : reg->top_h;
    tex_top_h += part->y_offset;

    // -AJA- 1999/09/25: Better DDF control of side texture.
    if (reg->extrafloor->type & EXFL_SideUpper)
      part = &cur_seg->sidedef->top;
    else if (reg->extrafloor->type & EXFL_SideLower)
      part = &cur_seg->sidedef->bottom;
  }
 
  // ignore missing textures (potentially causing HOM)
  if (! part->image)
    return;

  c = MIN((*c1), top);
  f = MAX((*f1), bottom);

  if (f >= c)
    return;
  
  RGL_DrawWall(dfloor, c, f, tex_top_h, part, 0, opaque, x_offset);

  // update current f1 & c1 heights
  if (part->translucency < 1.0)
    return;

  if (top > (*c1))
    (*c1) = bottom;
  
  if (bottom < (*f1))
    (*f1) = top;
}
  

//
// RGL_BuildWalls
//
// Analyses floor/ceiling heights, and add corresponding walls/floors
// to the drawfloor.  Returns true if the whole region was "solid".
//
static boolean_t RGL_BuildWalls(drawfloor_t *dfloor)
{
  float_t f0 = cur_sub->sector->floor.h;
  float_t c0 = cur_sub->sector->ceil.h;
  
  float_t f1 = dfloor->reg->f_h;
  float_t c1 = dfloor->reg->c_h;

  float_t f2, c2;

  vert_region_t *reg;
  sidepart_t *part;
  float_t tex_top_h;
  int line_flags;

#if (DEBUG >= 3)
    I_WriteDebug( "   BUILD WALLS %1.1f .. %1.1f\n", f1, c1);
#endif

  // minisegs never have any walls
  if (cur_seg->miniseg)
    return false;

  // handle TRANSLUCENT + THICK floors (a bit of a hack)
  if (dfloor->reg->extrafloor && dfloor->reg->sec_next &&
      (dfloor->reg->extrafloor->type & EXFL_Thick) &&
      (dfloor->reg->ceil->translucency <= 0.99))
  {
    c1 = dfloor->reg->top_h;
  }

  line_flags = cur_seg->linedef->flags;

  if (!backsector || !backsector->regions)
  {
    // one sided wall

    part = &cur_seg->sidedef->middle;

    // ignore missing textures (potentially causing HOM)
    if (! part->image)
      return false;
 
    tex_top_h = (line_flags & ML_DONTPEGBOTTOM) ?
        f0 + part->image->actual_h : c0;

    RGL_DrawWall(dfloor, c1, f1, tex_top_h + part->y_offset, 
        part, 0, true, part->x_offset);

    return true;
  }

  // --- two sided wall ---

  // first handle the lower part
  RGL_BuildWallSection(dfloor, NULL, backsector->floor.h, f0,
      line_flags, f0, c0, &f1, &c1);
  
  // then handle the other parts, including the upper part
  for (reg=backsector->regions; reg && f1 < c1; reg=reg->sec_next)
  {
    RGL_BuildWallSection(dfloor, reg, 
      reg->sec_next ? reg->top_h : c0, reg->c_h,
      line_flags, f0, c0, &f1, &c1);
  }
  
  // lastly handle the mid-masked texture, if any
  if (cur_seg->sidedef->middle.image == NULL)
  {
    // -AJA- hack for transparent doors (this test would normally be
    // above this block, not inside it).
    if (f1 >= c1)
      return true;

    return false;
  }

  part = &cur_seg->sidedef->middle;

  // ignore invisible walls
  if (part->translucency < 0.01)
    return false;

  f1 = MAX(f0, backsector->floor.h);
  c1 = MIN(c0, backsector->ceil.h);

  DEV_ASSERT2(part->image);
  if (line_flags & ML_DONTPEGBOTTOM)
  {
    f2 = f1 + part->y_offset;
    c2 = f2 + part->image->actual_h;
  }
  else
  {
    c2 = c1 + part->y_offset;
    f2 = c2 - part->image->actual_h;
  }

  tex_top_h = c2;

  // -AJA- more hacks for transparent doors
  if (cur_lower_invis) f1 = f0;
  if (cur_upper_invis) c1 = c0;

  f2 = MAX(MAX(f2, f1), dfloor->reg->f_h);
  c2 = MIN(MIN(c2, c1), dfloor->reg->c_h);

  if (c2 > f2)
    RGL_DrawWall(dfloor, c2, f2, tex_top_h, part, 1, false,
        part->x_offset);

  return false;
}

//
// RGL_WalkWall
//
static void RGL_WalkWall(seg_t *seg)
{
  drawfloor_t *dfloor;

  cur_seg = seg;

#if (DEBUG >= 2)
    I_WriteDebug("   DRAW SEG %p\n", seg);
#endif

  // mark the segment on the automap
  if (! seg->miniseg)
    seg->linedef->flags |= ML_Mapped;

  // --- handle each floor ---
  
  frontsector = seg->front_sub->sector;
  backsector  = NULL;

  if (seg->back_sub)
    backsector = seg->back_sub->sector;

  cur_upper_sky = cur_lower_sky = false;
  cur_upper_invis = cur_lower_invis = false;

  if (backsector && IS_SKY(frontsector->ceil) && IS_SKY(backsector->ceil))
    cur_upper_sky = true;

  if (backsector && IS_SKY(frontsector->floor) && IS_SKY(backsector->floor))
    cur_lower_sky = true;

  // -AJA- hack to allow transparent doors
  if (!seg->miniseg && backsector)
  {
    cur_lower_invis = (seg->sidedef->bottom.image == NULL) &&
        (backsector->floor.h != frontsector->floor.h);

    cur_upper_invis = (seg->sidedef->top.image == NULL) &&
        (backsector->ceil.h != frontsector->ceil.h);
  }

  for (dfloor=cur_sub->floors; dfloor; dfloor=dfloor->next)
  {
    RGL_BuildWalls(dfloor);
  }
}


//
// RGL_WalkSeg
//
// Visit a single seg of the subsector, and for one-sided lines update
// the 1D occlusion buffer.
//
static void RGL_WalkSeg(seg_t *seg)
{
  angle_t angle1, angle2;
  angle_t span, tspan1, tspan2;

  seg->visible = false;

  // compute distances
  {
    float_t tx1 = seg->v1->x - viewx;
    float_t ty1 = seg->v1->y - viewy;
    float_t tx2 = seg->v2->x - viewx;
    float_t ty2 = seg->v2->y - viewy;

    seg->tx1 = tx1 * viewsin - ty1 * viewcos;
    seg->tz1 = tx1 * viewcos + ty1 * viewsin;

    seg->tx2 = tx2 * viewsin - ty2 * viewcos;
    seg->tz2 = tx2 * viewcos + ty2 * viewsin;
  }

  if (seg->miniseg)
    return;

  angle1 = R_PointToAngle(viewx, viewy, seg->v1->x, seg->v1->y);
  angle2 = R_PointToAngle(viewx, viewy, seg->v2->x, seg->v2->y);

#if (DEBUG >= 3)
    I_WriteDebug( "INIT ANGLE1 = %1.2f  ANGLE2 = %1.2f\n", 
        ANG_2_FLOAT(angle1), ANG_2_FLOAT(angle2));
#endif

  // Clip to view edges.
  // -ES- 1999/03/20 Replaced clipangle with clipscope/leftclipangle/rightclipangle

  span = angle1 - angle2;

  // back side ?
  if (span >= ANG180)
    return;

  angle1 -= viewangle;
  angle2 -= viewangle;

#if (DEBUG >= 3)
    I_WriteDebug( "ANGLE1 = %1.2f  ANGLE2 = %1.2f\n", 
        ANG_2_FLOAT(angle1), ANG_2_FLOAT(angle2));
#endif

  tspan1 = angle1 - rightclipangle;
  tspan2 = leftclipangle - angle2;

  if (tspan1 > clipscope)
  {
    // Totally off the left edge?
    if (tspan2 >= ANG180)
      return;

    angle1 = leftclipangle;
  }

  if (tspan2 > clipscope)
  {
    // Totally off the left edge?
    if (tspan1 >= ANG180)
      return;

    angle2 = rightclipangle;
  }

#if (DEBUG >= 3)
    I_WriteDebug( "CLIPPED ANGLE1 = %1.2f  ANGLE2 = %1.2f\n", 
        ANG_2_FLOAT(angle1), ANG_2_FLOAT(angle2));
#endif

  // The seg is in the view range,
  // but not necessarily visible.

#if (DEBUG >= 2)
    I_WriteDebug( "  %sSEG %p (%1.1f, %1.1f) -> (%1.1f, %1.1f)\n",
        seg->miniseg ? "MINI" : "", seg, seg->v1->x, seg->v1->y, 
        seg->v2->x, seg->v2->y);
#endif
 
  // check if visible

#if 1
  if (span > ANG1 && RGL_1DOcclusionTest(angle2, angle1))
  {
    return;
  }
#endif

  seg->visible = true;

  seg->angle1 = angle1;
  seg->angle2 = angle2;

  if (IS_SKY(cur_sub->sector->floor) || IS_SKY(cur_sub->sector->ceil))
  {
    RGL_MarkSky();
  }

  // only 1 sided walls affect the 1D occlusion buffer
  // FIXME: may need a hack for transparent doors
  if (!seg->backsector ||
      seg->backsector->ceil.h <= seg->frontsector->floor.h ||
      seg->backsector->floor.h >= seg->frontsector->ceil.h)
  {
    RGL_1DOcclusionSet(angle2, angle1);
  }
}

//
// RGL_CheckBBox
//
// Checks BSP node/subtree bounding box.
// Returns true if some part of the bbox might be visible.
//
// Placed here to be close to RGL_WalkSeg(), which has similiar angle
// clipping stuff in it.
//
extern int checkcoord[12][4];

boolean_t RGL_CheckBBox(float_t *bspcoord)
{
  int boxx, boxy;
  int boxpos;

  float_t x1, y1, x2, y2;

  angle_t angle1, angle2;
  angle_t span, tspan1, tspan2;

  // Find the corners of the box
  // that define the edges from current viewpoint.
  if (viewx <= bspcoord[BOXLEFT])
    boxx = 0;
  else if (viewx < bspcoord[BOXRIGHT])
    boxx = 1;
  else
    boxx = 2;

  if (viewy >= bspcoord[BOXTOP])
    boxy = 0;
  else if (viewy > bspcoord[BOXBOTTOM])
    boxy = 1;
  else
    boxy = 2;

  boxpos = (boxy << 2) + boxx;

  if (boxpos == 5)
    return true;

  x1 = bspcoord[checkcoord[boxpos][0]];
  y1 = bspcoord[checkcoord[boxpos][1]];
  x2 = bspcoord[checkcoord[boxpos][2]];
  y2 = bspcoord[checkcoord[boxpos][3]];

  // check clip list for an open space
  angle1 = R_PointToAngle(viewx, viewy, x1, y1) - viewangle;
  angle2 = R_PointToAngle(viewx, viewy, x2, y2) - viewangle;

  span = angle1 - angle2;

  // Sitting on a line?
  if (span >= ANG180)
    return true;

  // -ES- 1999/03/20 Replaced clipangle with clipscope/leftclipangle/rightclipangle

  tspan1 = angle1 - rightclipangle;
  tspan2 = leftclipangle - angle2;

  if (tspan1 > clipscope)
  {
    // Totally off the left edge?
    if (tspan2 >= ANG180)
      return false;

    angle1 = leftclipangle;
  }

  if (tspan2 > clipscope)
  {
    // Totally off the right edge?
    if (tspan1 >= ANG180)
      return false;

    angle2 = rightclipangle;
  }

  return ! RGL_1DOcclusionTest(angle2, angle1);
}

//
// RGL_DrawPlane
//
#define MAX_PLVERT  64

static void RGL_DrawPlane(drawfloor_t *dfloor, 
    plane_info_t *info, int face_dir)
{
  seg_t *seg;
  float_t n_z;

  float_t p_cos = M_Cos(info->angle);
  float_t p_sin = M_Sin(info->angle);

  boolean_t blended;

  GLuint tex_id;
  const cached_image_t *cim;
  local_gl_vert_t *vert, *orig;

  int num_vert, i;
  int L_r[MAX_PLVERT], L_g[MAX_PLVERT], L_b[MAX_PLVERT];

  region_properties_t *props = info->override_p ?
      info->override_p : dfloor->reg->p;

  int lit_Nom = RGL_Light(props->lightlevel);
  const colourmap_t *colmap = props->colourmap;
  float_t c_r, c_g, c_b;

  float_t trans = info->translucency;

  // ignore invisible planes
  if (trans < 0.01)
    return;
 
  // ignore sky
  if (IS_SKY(*info))
    return;

  // ignore non-facing planes
  if ((viewz > info->h) != (face_dir > 0))
    return;

  // ignore non-solid planes in solid_mode (& vice versa)
  blended = (trans <= 0.99) ? true : false;
  if (blended == solid_mode)
    return;

  // ignore dud regions (floor >= ceiling)
  if (dfloor->reg->f_h > dfloor->reg->c_h)
    return;

  // ignore empty subsectors
  if (cur_sub->segs == NULL)
    return;

  // count number of actual vertices
  for (seg=cur_sub->segs, num_vert=0; seg; seg=seg->sub_next, num_vert++)
  {
    /* nothing here */
  }

  if (num_vert > MAX_PLVERT)
  {
    num_vert = MAX_PLVERT;
  }
  
  V_GetColmapRGB(colmap, &c_r, &c_g, &c_b, false);

  for (i=0; i < num_vert; i++)
  {
    L_r[i] = lit_Nom * c_r;
    L_g[i] = lit_Nom * c_g;
    L_b[i] = lit_Nom * c_b;
  }

  DEV_ASSERT2(info->image);
  cim = W_ImageCache(info->image, IMG_OGL, 0, true);
  tex_id = W_ImageGetOGL(cim);
  
  vert = orig = RGL_BeginUnit(GL_POLYGON, MAX_PLVERT, tex_id, false, blended);

  // Note: normally this would be wrong, since we're using the GL
  // texture ID later on (after W_ImageDone).  The W_LockImagesOGL
  // call saves us though.
  W_ImageDone(cim);

  //
  //  Dynamic Lighting Stuff
  //

  if (use_dlights)
  {
    drawthing_t *dl;
    float_t wx[MAX_PLVERT], wy[MAX_PLVERT], wz[MAX_PLVERT];

    for (seg=cur_sub->segs, i=0; seg && (i < MAX_PLVERT); 
         seg=seg->sub_next, i++)
    {
      wx[i] = seg->v1->x;
      wy[i] = seg->v1->y;
      wz[i] = info->h;
    }

    for (dl=dfloor->dlights; dl; dl=dl->next)
    {
      // light behind the plane ?    
      if ((dl->tz > info->h) != (face_dir > 0))
        continue;
       
      R2_AddColourDLights(num_vert, L_r, L_g, L_b, wx, wy, wz, dl->mo);
    }
  }

  n_z = (viewz > info->h) ? 1.0 : -1.0;

  for (seg=cur_sub->segs, i=0; seg && (i < MAX_PLVERT); 
       seg=seg->sub_next, i++)
  {
    float_t x = seg->v1->x;
    float_t y = seg->v1->y;
    float_t z = info->h;

    float_t rx = x + info->x_offset;
    float_t ry = y + info->y_offset;

    float_t tx = (rx * p_cos + ry * p_sin) / info->scale;
    float_t ty = (ry * p_cos - rx * p_sin) / info->scale;

    SET_COLOR(LT_RED(L_r[i]), LT_GRN(L_g[i]), LT_BLU(L_b[i]), trans);
    SET_TEXCOORD(tx, ty);
    SET_NORMAL(0.0, 0.0, n_z);
    SET_EDGE_FLAG(GL_TRUE);
    SET_VERTEX(x, y, z);

    vert++;
  }

  RGL_EndUnit(vert - orig);
}

//
// RGL_WalkThing
//
// Visit a single thing that exists in the current subsector.
//
static void RGL_WalkThing(mobj_t *mo)
{
  drawthing_t *dthing;
  
  float_t tr_x;
  float_t tr_y;

  float_t pos1, pos2;
  float_t tx, tx1, tx2;
  float_t tz;

  float_t gzb, gzt;

  float_t xscale;
  float_t yscale;
  float_t dist_scale;
  int clip_vert = 0;

  int sprite_height;
  int top_offset;

  spritedef_t *sprite;
  spriteframe_t *spr_frame;
  int spr_lump;
  boolean_t spr_flip;
  boolean_t spr_bright;

  // ignore invisible things
  if (mo->visibility == INVISIBLE)
      return;

  // transform the origin point
  tr_x = mo->x - viewx;
  tr_y = mo->y - viewy;

  tz = tr_x * viewcos + tr_y * viewsin;

  // thing is behind view plane?
  if (tz < MINZ)
    return;

  // -ES- 1999/03/14 Use distunit for y and x scale.
  CHECKVAL(tz);
  xscale = x_distunit / tz;
  yscale = y_distunit / tz;
  dist_scale = yscale;

  tx = tr_x * viewsin - tr_y * viewcos;

  // too far off the side?
  // -ES- 1999/03/13 Fixed clipping to work with large FOVs (up to 176 deg)
  // rejects all sprites where angle>176 deg (arctan 32), since those
  // sprites would result in overflow in future calculations
  if (fabs(tx) / 32 > tz)
    return;
  
  R2_GetThingSprite(mo, &sprite, &spr_frame, &spr_lump, &spr_flip,
      &spr_bright);

  // calculate edges of the shape
  pos1 = - spriteoffset[spr_lump] * mo->info->xscale;
  pos2 = pos1 + spritewidth[spr_lump] * mo->info->xscale;
  
  tx1 = tx + pos1;
  tx2 = tx + pos2;

  xscale *= mo->info->xscale;
  yscale *= mo->info->yscale;

  // fix for sprites that sit wrongly into the floor
  sprite_height = spriteheight[spr_lump];
  top_offset = spritetopoffset[spr_lump];

  gzt = mo->z + top_offset * mo->info->yscale;
  gzb = gzt - sprite_height * mo->info->yscale;

  // fix for sprites that sit wrongly into the floor/ceiling

  if (mo->flags & MF_FUZZY)
  {
    clip_vert = -1;
  }
  else if (sprite_kludge==0 && gzb < mo->floorz)
  {
    // explosion ?
    if (mo->info->flags & MF_MISSILE)
      clip_vert = +1;
    else
    {
      gzt += mo->floorz - gzb;
      gzb = mo->floorz;
    }

    if (gzb >= gzt)
      return;
  }
  else if (sprite_kludge==0 && gzt > mo->ceilingz)
  {
    // explosion ?
    if (mo->info->flags & MF_MISSILE)
      clip_vert = +1;
    else
    {
      gzb += gzt - mo->ceilingz;
      gzt = mo->ceilingz;
    }

    if (gzb >= gzt)
      return;
  }

  // create new draw thing

  dthing = R2_GetDrawThing();
  R2_CommitDrawThing(1);

  dthing->mo = mo;
  dthing->props = cur_sub->floors->reg->p;
  dthing->clip_vert = clip_vert;
  dthing->clipped_left = dthing->clipped_right = false;

  dthing->patch  = spritelist[spr_lump];
  dthing->flip   = spr_flip;
  dthing->bright = spr_bright;
  dthing->is_shadow = false;
  
  dthing->xscale = 1.0;
  dthing->yscale = 1.0;
  dthing->dist_scale = dist_scale;

  dthing->tx = tx;
  dthing->tz = tz;
  dthing->tx1 = tx1;
  dthing->tx2 = tx2;

  dthing->top = gzt;
  dthing->bottom = gzb;

  dthing->left_dx  = pos1 *  viewsin;
  dthing->left_dy  = pos1 * -viewcos;
  dthing->right_dx = pos2 *  viewsin;
  dthing->right_dy = pos2 * -viewcos;
  
  // translation support
  if (mo->info->palremap)
    dthing->trans_table = V_GetTranslationTable(mo->info->palremap);
  else
    dthing->trans_table = NULL;

  R2_ClipSpriteVertically(cur_sub, dthing);

  // create shadow
  if (level_flags.shadows && mo->info->shadow_trans > 0 &&
      mo->floorz < viewz && ! IS_SKY(mo->subsector->sector->floor))
  {
    drawthing_t *dshadow = R2_GetDrawThing();
    R2_CommitDrawThing(1);

    dshadow[0] = dthing[0];

    dshadow->is_shadow = true;
    dshadow->clip_vert = -1;
    dshadow->trans_table = NULL;
    dshadow->tz -= 1.1;

    // shadows are 1/4 the height
    dshadow->yscale *= 4.0;

    gzb = mo->floorz;
    gzt = gzb + sprite_height / 4.0 * mo->info->yscale;

    dshadow->top = gzt;
    dshadow->bottom = gzb;

    R2_ClipSpriteVertically(cur_sub, dshadow);
  }
}

//
// RGL_DrawThing
//
void RGL_DrawThing(drawfloor_t *dfloor, drawthing_t *dthing)
{
  int w, h;
  float_t right, bottom;

  float_t dx, dy;
  float_t tex_x1=0, tex_y1=0;
  float_t tex_x2=0, tex_y2=0;

  boolean_t blended;
  boolean_t texturing;

  local_gl_vert_t *vert, *orig;
  GLuint tex_id;

  float_t x1b, y1b, z1b, x1t, y1t, z1t;
  float_t x2b, y2b, z2b, x2t, y2t, z2t;

  int i, L_r[4], L_g[4], L_b[4];

  int lit_Nom = dthing->bright ? 255 : 
      RGL_Light(dthing->props->lightlevel);
  const colourmap_t *colmap = dthing->props->colourmap;
  float_t c_r, c_g, c_b;

  int fuzzy = (dthing->mo->flags & MF_FUZZY);
  float_t trans = fuzzy ? FUZZY_TRANS : dthing->mo->visibility;

  V_GetColmapRGB(colmap, &c_r, &c_g, &c_b, false);

  if (dthing->is_shadow)
  {
    for (i=0; i < 4; i++)
      L_r[i] = L_g[i] = L_b[i] = 0;

    trans = dthing->mo->visibility * 
        PERCENT_2_FLOAT(dthing->mo->info->shadow_trans);

    dx = viewcos * 2.3;
    dy = viewsin * 2.3;
  }
  else
  {
    for (i=0; i < 4; i++)
    {
      L_r[i] = lit_Nom * c_r;
      L_g[i] = lit_Nom * c_g;
      L_b[i] = lit_Nom * c_b;
    }

    dx = dy = 0;
  }

  tex_id = dthing->patch;

  if (tex_id >= MAX_TEX)
    return;

  if (conv_sprites[tex_id].status == TEX_EMPTY)
    RGL_Sprite(tex_id);

  w = conv_sprites[tex_id].width;
  h = conv_sprites[tex_id].height;
  right  = conv_sprites[tex_id].right;
  bottom = conv_sprites[tex_id].bottom;

  texturing = (conv_sprites[tex_id].status == TEX_OK);

  // NOTE: code expects new image system, for now return if no tex
  if (! texturing)
    return;

  x1b = x1t = dthing->mo->x + dthing->left_dx;
  y1b = y1t = dthing->mo->y + dthing->left_dy;
  x2b = x2t = dthing->mo->x + dthing->right_dx;
  y2b = y2t = dthing->mo->y + dthing->right_dy;

  z1b = z2b = dthing->bottom;
  z1t = z2t = dthing->top;

  tex_x1 = 0;
  tex_x2 = right;

  tex_y1 = 0;
  tex_y2 = tex_y1 + (z1t - z1b) * dthing->yscale;

  CHECKVAL(h);
  tex_y1 /= h;
  tex_y2 /= h;
  
  if (dthing->flip)
  {
    float_t temp = tex_x2;
    tex_x1 = right - tex_x1;
    tex_x2 = right - temp;
  }

  //
  //  Dynamic Lighting Stuff
  //

  if (use_dlights && !dthing->is_shadow && !fuzzy)
  {
    drawthing_t *dl;
    float_t wx[4], wy[4], wz[4];

    wx[0] = x1b;  wy[0] = y1b;  wz[0] = z1b;
    wx[1] = x1t;  wy[1] = y1t;  wz[1] = z1t;
    wx[2] = x2t;  wy[2] = y2t;  wz[2] = z2t;
    wx[3] = x2b;  wy[3] = y2b;  wz[3] = z2b;

    for (dl=dfloor->dlights; dl; dl=dl->next)
    {
      if (dl->mo == dthing->mo)
        continue;

      R2_AddColourDLights(4, L_r, L_g, L_b, wx, wy, wz, dl->mo);
    }
  }
  
  //
  // Special FUZZY effect
  //
  if (!dthing->is_shadow && fuzzy)
  {
    float_t range_x = fabs(dthing->right_dx - dthing->left_dx) / 12.0;
    float_t range_y = fabs(dthing->right_dy - dthing->left_dy) / 12.0;
    float_t range_z = fabs(z1t - z1b) / 24.0 / 2;

    float_t bl = sin(rgl_currtimeval / 5.0);
    float_t tl = sin(rgl_currtimeval / 11.0);
    float_t tr = sin(rgl_currtimeval / 7.0);
    float_t br = sin(rgl_currtimeval / 13.0);

    float_t zbl = 1.0 + cos(rgl_currtimeval / 5.0);
    float_t ztl = 1.0 + cos(rgl_currtimeval / 11.0);
    float_t ztr = 1.0 + cos(rgl_currtimeval / 7.0);
    float_t zbr = 1.0 + cos(rgl_currtimeval / 13.0);
    
    x1b += bl * range_x; y1b += bl * range_y; z1b += zbl * range_z;
    x1t += tl * range_x; y1t += tl * range_y; z1t += ztl * range_z;
    x2t += tr * range_x; y2t += tr * range_y; z2t += ztr * range_z;
    x2b += br * range_x; y2b += br * range_y; z2b += zbr * range_z;

    for (i=0; i < 4; i++)
    {
      L_r[i] = L_g[i] = L_b[i] = 0;
    }
  }

  // Blended sprites, even if opaque (trans > 0.99), have nicer
  // edges.
  blended = true;

  vert = orig = RGL_BeginUnit(GL_QUADS, 4, conv_sprites[tex_id].id,
      true, blended);

  SET_COLOR(LT_RED(L_r[0]), LT_GRN(L_g[0]), LT_BLU(L_b[0]), trans);
  SET_TEXCOORD(tex_x1, bottom - tex_y2);
  SET_NORMAL(-viewcos, -viewsin, 0.0);
  SET_EDGE_FLAG(GL_TRUE);
  SET_VERTEX(x1b+dx, y1b+dy, z1b);
  vert++;
  
  SET_COLOR(LT_RED(L_r[1]), LT_GRN(L_g[1]), LT_BLU(L_b[1]), trans);
  SET_TEXCOORD(tex_x1, bottom - tex_y1);
  SET_NORMAL(-viewcos, -viewsin, 0.0);
  SET_EDGE_FLAG(GL_TRUE);
  SET_VERTEX(x1t+dx, y1t+dy, z1t);
  vert++;
  
  SET_COLOR(LT_RED(L_r[2]), LT_GRN(L_g[2]), LT_BLU(L_b[2]), trans);
  SET_TEXCOORD(tex_x2, bottom - tex_y1);
  SET_NORMAL(-viewcos, -viewsin, 0.0);
  SET_EDGE_FLAG(GL_TRUE);
  SET_VERTEX(x2t+dx, y2t+dy, z2t);
  vert++;
  
  SET_COLOR(LT_RED(L_r[3]), LT_GRN(L_g[3]), LT_BLU(L_b[3]), trans);
  SET_TEXCOORD(tex_x2, bottom - tex_y2);
  SET_NORMAL(-viewcos, -viewsin, 0.0);
  SET_EDGE_FLAG(GL_TRUE);
  SET_VERTEX(x2b+dx, y2b+dy, z2b);
  vert++;

  RGL_EndUnit(vert - orig);

#if 0 //!!! (HALOS)  // EXPERIMENTAL!
  if (!thing->bright || !level_flags.halos)
    return;

  tex_id = SPR_HALO;

  if (conv_sprites[tex_id].status == TEX_EMPTY)
    RGL_Sprite(tex_id);

  if (conv_sprites[tex_id].status != TEX_OK)
    return;

  w = conv_sprites[tex_id].width;
  h = conv_sprites[tex_id].height;
  right  = conv_sprites[tex_id].right;
  bottom = conv_sprites[tex_id].bottom;

  tex_x1 = thing->mo->x - thing->ixscale * 32 *  viewsin;
  tex_y1 = thing->mo->y - thing->ixscale * 32 * -viewcos;

  tex_x2 = thing->mo->x + thing->ixscale * 32 *  viewsin;
  tex_y2 = thing->mo->y + thing->ixscale * 32 * -viewcos;

  tex_top_h = thing->mo->z + thing->mo->height*0.75 + thing->ixscale * 32;
  tex_bot_h = thing->mo->z + thing->mo->height*0.75 - thing->ixscale * 32;

  glEnable(GL_BLEND);
  glEnable(GL_TEXTURE_2D);

  glBindTexture(GL_TEXTURE_2D, conv_sprites[tex_id].id);

  glColor4f(1.0, 0.6, 0.0, 0.5);

  glBegin(GL_QUADS);
  
  glTexCoord2f(0.0, bottom);
  glVertex3f(tex_x1,  tex_y1, tex_bot_h);
  
  glTexCoord2f(0.0, 0.0);
  glVertex3f(tex_x1,  tex_y1, tex_top_h);
  
  glTexCoord2f(right, 0.0);
  glVertex3f(tex_x2, tex_y2, tex_top_h);
  
  glTexCoord2f(right, bottom);
  glVertex3f(tex_x2, tex_y2, tex_bot_h);
  
  glEnd();
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_ALPHA_TEST);
  glDisable(GL_BLEND);
#endif
}

//
// RGL_DrawSortThings
//
static drawthing_t **mapdrawthings = NULL;
static int mapdrawthing_num = 0;

static void RGL_DrawSortThings(drawfloor_t *dfloor)
{
  drawthing_t *dthing;

  int i, count=0;

  for (dthing=dfloor->things; dthing; dthing=dthing->next)
    count++;

  if (count == 0)
    return;

  // maybe reallocate array
  if (count > mapdrawthing_num)
  {
    Z_Resize(mapdrawthings, drawthing_t *, count);
    mapdrawthing_num = count;
  }

  // fill array
  count=0;

  for (dthing=dfloor->things; dthing; dthing=dthing->next)
    mapdrawthings[count++] = dthing;

  // sort array
  #define CMP(a,b)  ((a)->tz > (b)->tz)
  QSORT(drawthing_t *, mapdrawthings, count, CUTOFF);
  #undef CMP

  // draw array
  for (i=0; i < count; i++)
    RGL_DrawThing(dfloor, mapdrawthings[i]);
}

static void LightUpPlayerWeapon(player_t *p, drawfloor_t *dfloor)
{
  drawthing_t *dl;
  float_t wx[4], wy[4], wz[4];

  float_t x, y, z;
  float_t dx, dy, fx, fy;

  x = p->mo->x;
  y = p->mo->y;
  z = p->mo->z + p->mo->height *
      PERCENT_2_FLOAT(p->mo->info->shotheight);

  dx = viewcos * p->mo->radius;
  dy = viewsin * p->mo->radius;

  fx = ( dy) * 2.0;
  fy = (-dx) * 2.0;

  wx[0] = x-dx;     wy[0] = y-dy;     wz[0] = z;
  wx[1] = x-dx+fx;  wy[1] = y-dy+dy;  wz[1] = z;
  wx[2] = x+dx+fx;  wy[2] = y+dy+fy;  wz[2] = z;
  wx[3] = x+dx;     wy[3] = y+dy;     wz[3] = z;

  for (dl=dfloor->dlights; dl; dl=dl->next)
  {
    R2_AddColourDLights(4, rgl_weapon_r, rgl_weapon_g, rgl_weapon_b, 
        wx, wy, wz, dl->mo);
  }
}

//
// RGL_WalkSubsector
//
// Visit a subsector, and collect information, such as where the
// walls, planes (ceilings & floors) and things need to be drawn.
//
static void RGL_WalkSubsector(int num)
{
  subsector_t *dsub = &subsectors[num];
  vert_region_t *reg;
  seg_t *seg;
  mobj_t *mo;

  drawfloor_t *dfloor;
  drawfloor_t *tail;

#if (DEBUG >= 1)
    I_WriteDebug( "\nVISITING SUBSEC %d\n\n", num);
#endif

  cur_sub = dsub;

  dsub->ranges = NULL;
  dsub->floors = dsub->z_floors = NULL;
  dsub->raw_things = NULL;

  // add in each floor
  for (reg=cur_sub->sector->regions; reg; reg=reg->sec_next)
  {
    dfloor = R2_GetDrawFloor();
    R2_CommitDrawFloor(1);

    dfloor->next = NULL;
    dfloor->reg  = reg;

    dfloor->walls  = NULL;
    dfloor->planes = NULL;
    dfloor->things = NULL;
    dfloor->dlights = NULL;

    // link it in
    // (order is very important)

    if (dsub->floors == NULL || reg->f_h > viewz)
    {
      // add to head
      dfloor->next = dsub->floors;

      dsub->floors = dfloor;
    }
    else
    {
      // add to tail
      for (tail=dsub->floors; tail->next; tail=tail->next)
      { /* nothing here */ }

      tail->next = dfloor;
    }

    // add to tail of height order list (for sprite clipping)
    for (tail=dsub->z_floors; tail && tail->z_next; tail=tail->z_next)
    { /* nothing here */ }

    dfloor->z_next = NULL;
    dfloor->z_prev = tail;

    if (tail)
      tail->z_next = dfloor;
    else
      dsub->z_floors = dfloor;

    if (use_dlights && blocklights)
    {
      R2_FindDLights(cur_sub, dfloor);
    
      // light up the player's weapon
      if (reg == view_vertregion)
        LightUpPlayerWeapon(consoleplayer, dfloor);
    }
  }

  // handle each sprite in the subsector.  Must be done before walls,
  // since the wall code will update the 1D occlusion buffer.
  
  for (mo=cur_sub->thinglist; mo; mo=mo->snext)
  {
    RGL_WalkThing(mo);
  }

  // clip 1D occlusion buffer.
  for (seg=dsub->segs; seg; seg=seg->sub_next)
  {
    RGL_WalkSeg(seg);
  }

  // add drawsub to list
  // (add to head, thus the eventual order is furthest -> closest)
  
  dsub->rend_next = drawsubs_head;
  dsub->rend_prev = NULL;
  
  if (drawsubs_head)
    drawsubs_head->rend_prev = dsub;
  else
    drawsubs_tail = dsub;

  drawsubs_head = dsub;
}


//
// RGL_DrawSubsector
//
static void RGL_DrawSubsector(subsector_t *dsub)
{
  drawfloor_t *dfloor;
  seg_t *seg;

#if (DEBUG >= 1)
    I_WriteDebug("\nREVISITING SUBSEC %d\n\n", (int)(dsub - subsectors));
#endif

  cur_sub = dsub;

  // handle each seg in the subsector
  for (seg=dsub->segs; seg; seg=seg->sub_next)
  {
    if (seg->visible)
      RGL_WalkWall(seg);
  }

  // handle each floor, drawing planes and things
  for (dfloor=dsub->floors; dfloor; dfloor=dfloor->next)
  {
    RGL_DrawPlane(dfloor, dfloor->reg->floor, +1);
    RGL_DrawPlane(dfloor, dfloor->reg->ceil,  -1);

    if (solid_mode)
      continue;

    RGL_DrawSortThings(dfloor);
  }
}

//
// RGL_WalkBSPNode
//
// Walks all subsectors below a given node, traversing subtree
// recursively, collecting information.  Just call with BSP root.
//
static void RGL_WalkBSPNode(int bspnum)
{
  node_t *node;
  int side;

  // Found a subsector?
  if (bspnum & NF_SUBSECTOR)
  {
    RGL_WalkSubsector(bspnum & (~NF_SUBSECTOR));
    return;
  }

  node = &nodes[bspnum];

  // Decide which side the view point is on.
  side = P_PointOnDivlineSide(viewx, viewy, &node->div);

#if (DEBUG >= 2)
    I_WriteDebug( "NODE %d (%1.1f, %1.1f) -> (%1.1f, %1.1f)  SIDE %d\n",
        bspnum, node->div.x, node->div.y, node->div.x +
        node->div.dx, node->div.y + node->div.dy, side);
#endif

  // Recursively divide front space.
  if (RGL_CheckBBox(node->bbox[side]))
    RGL_WalkBSPNode(node->children[side]);

  // Recursively divide back space.
  if (RGL_CheckBBox(node->bbox[side ^ 1]))
    RGL_WalkBSPNode(node->children[side ^ 1]);
}

//
// RGL_RenderTrueBSP
//
// OpenGL BSP rendering.  Initialises all structures, then walks the
// BSP tree collecting information, then renders each subsector:
// firstly front to back (drawing all solid walls & planes) and then
// from back to front (drawing everything else, sprites etc..).
//
void RGL_RenderTrueBSP(void)
{
  subsector_t *cur;
  int i;

  rgl_currtimeval = I_GetTime();

  // compute the 1D projection of the view angle
  {
    float_t k, d;

    // k is just the mlook angle (in radians)
    k = fabs(atan(viewvertangle));

    if (k >= M_PI/4)
      oned_side_angle = ANG90;
    else
    {
      // d is just the distance horizontally forward from the eye to
      // the top/bottom edge of the view rectangle.
      d = M_ROOT2 * sin(M_PI/2 - (k + M_PI/4));

      oned_side_angle = ANG90 - M_ATan(d);
    }
  }

  // setup clip angles
  {
    leftclipangle  = 0 + oned_side_angle;
    rightclipangle = 0 - oned_side_angle;
    clipscope = leftclipangle - rightclipangle;
  }

  // clear extra light on player's weapon
  for (i=0; i < 4; i++)
  {
    rgl_weapon_r[i] = rgl_weapon_g[i] = rgl_weapon_b[i] = 0;
  }

  R2_ClearBSP();
  RGL_1DOcclusionClear(0 - oned_side_angle, 0 + oned_side_angle);

  drawsubs_head = drawsubs_tail = NULL;

  // walk the bsp tree
  RGL_WalkBSPNode(root_node);

  // handle powerup effects
  RGL_RainbowEffect(consoleplayer);

  // NOTE: this call will be unneeded later.
  RGL_SetupMatrices2D();

  RGL_DrawSky();

  RGL_SetupMatrices3D();

  // enable depth buffer, and clear it
  glEnable(GL_DEPTH_TEST);
  glClear(GL_DEPTH_BUFFER_BIT);

  // make sure we don't delete any GL textures
  W_LockImagesOGL();

  // draw all solid walls and planes
  solid_mode = true;
  RGL_StartUnits(solid_mode);

  for (cur=drawsubs_tail; cur; cur=cur->rend_prev)
    RGL_DrawSubsector(cur);
    
  RGL_FinishUnits();
  
  // draw all sprites and masked/translucent walls/planes
  solid_mode = false;
  RGL_StartUnits(solid_mode);
  
  for (cur=drawsubs_head; cur; cur=cur->rend_next)
    RGL_DrawSubsector(cur);

  RGL_FinishUnits();
  glDisable(GL_DEPTH_TEST);

  W_UnlockImagesOGL();

  // now draw 2D stuff like psprites, and add effects
  RGL_SetupMatrices2D();

  RGL_DrawPlayerSprites(consoleplayer);
  RGL_ColourmapEffect(consoleplayer);

  // NOTE: this call will move for layer system (should be topmost).
  RGL_PaletteEffect(consoleplayer);
 
#if (DEBUG >= 3) 
    I_WriteDebug( "\n\n");
#endif
}


void RGL_RenderScene(int x, int y, int w, int h, vid_view_t *view)
{
  //!!!
  RGL_RenderTrueBSP();
}


#endif  // USE_GL
