//----------------------------------------------------------------------------
//  EDGE OpenGL Rendering (Skies)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2001  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------

// this conditional applies to the whole file
#ifdef USE_GL

#include "i_defs.h"

#include "am_map.h"
#include "dm_defs.h"
#include "dm_state.h"
#include "e_search.h"
#include "m_bbox.h"
#include "m_random.h"
#include "p_local.h"
#include "p_mobj.h"
#include "r_defs.h"
#include "r_main.h"
#include "r_plane.h"
#include "r_sky.h"
#include "r_state.h"
#include "r_things.h"
#include "r2_defs.h"
#include "rgl_defs.h"
#include "v_colour.h"
#include "v_ctx.h"
#include "v_res.h"
#include "w_textur.h"
#include "w_wad.h"
#include "z_zone.h"


#define DEBUG  0


#define RGB_RED(rgbcol)  ((float_t)((rgbcol >> 16) & 0xFF) / 255.0)
#define RGB_GRN(rgbcol)  ((float_t)((rgbcol >>  8) & 0xFF) / 255.0)
#define RGB_BLU(rgbcol)  ((float_t)((rgbcol      ) & 0xFF) / 255.0)

static boolean_t draw_sky = false;

//
// RGL_MarkSky
//
void RGL_MarkSky(void)
{
  draw_sky = true;
}

//
// RGL_SetupMatricesSky
//
// Setup the GL matrices for drawing a tiled sky.
//
#define Z_NEAR  1.0
#define Z_FAR   32000.0

void RGL_SetupMatricesTiledSky(void)
{
  glViewport(viewwindowx, SCREENHEIGHT - viewwindowy - viewwindowheight,
             viewwindowwidth, viewwindowheight);

  // calculate perspective matrix

  glMatrixMode(GL_PROJECTION);

  glLoadIdentity();
  glFrustum(rightslope, leftslope, bottomslope, topslope, Z_NEAR, Z_FAR);

  // calculate look-at matrix

  glMatrixMode(GL_MODELVIEW);

  glLoadIdentity();
  glRotatef(270.0 - atan(viewvertangle) * 180.0 / M_PI, 1.0, 0.0, 0.0);
  glRotatef(90.0 - ANG_2_FLOAT(viewangle), 0.0, 0.0, 1.0);
  glTranslatef(0, 0, 0);

  glDisable(GL_LIGHTING);
  glDisable(GL_COLOR_MATERIAL);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

static struct tilepoint_s
{
  float_t x, y, z, tx, ty, bx, by;
}
tilepoints[] =
{
 /* 0 */ { 0.000,1.000,0.000, 0.0000,1.0000, 0.0000,1.0000 },
 /* 1 */ { 0.000,0.938,0.348, 0.0312,0.9688, -0.0312,1.0312 },
 /* 2 */ { 0.067,0.998,0.000, 0.0344,1.0322, 0.0344,1.0322 },
 /* 3 */ { 0.000,0.875,0.484, 0.0625,0.9375, -0.0625,1.0625 },
 /* 4 */ { 0.067,0.935,0.348, 0.0658,1.0010, 0.0080,1.0677 },
 /* 5 */ { 0.141,0.990,0.000, 0.0757,1.0657, 0.0757,1.0657 },
 /* 6 */ { 0.000,0.812,0.583, 0.0938,0.9062, -0.0938,1.0938 },
 /* 7 */ { 0.067,0.872,0.484, 0.0973,0.9698, -0.0177,1.1040 },
 /* 8 */ { 0.143,0.927,0.348, 0.1080,1.0346, 0.0557,1.1059 },
 /* 9 */ { 0.225,0.974,0.000, 0.1252,1.0996, 0.1252,1.0996 },
 /* 10 */ { 0.000,0.750,0.661, 0.1250,0.8750, -0.1250,1.1250 },
 /* 11 */ { 0.067,0.810,0.583, 0.1289,0.9386, -0.0424,1.1410 },
 /* 12 */ { 0.144,0.863,0.484, 0.1404,1.0035, 0.0376,1.1473 },
 /* 13 */ { 0.227,0.910,0.348, 0.1589,1.0684, 0.1135,1.1442 },
 /* 14 */ { 0.316,0.949,0.000, 0.1838,1.1325, 0.1838,1.1325 },
 /* 15 */ { 0.000,0.688,0.726, 0.1562,0.8438, -0.1562,1.1562 },
 /* 16 */ { 0.068,0.747,0.661, 0.1605,0.9074, -0.0658,1.1790 },
 /* 17 */ { 0.145,0.799,0.583, 0.1730,0.9724, 0.0220,1.1904 },
 /* 18 */ { 0.230,0.844,0.484, 0.1930,1.0372, 0.1053,1.1907 },
 /* 19 */ { 0.320,0.881,0.348, 0.2197,1.1007, 0.1823,1.1808 },
 /* 20 */ { 0.414,0.910,0.000, 0.2517,1.1621, 0.2517,1.1621 },
 /* 21 */ { 0.000,0.625,0.781, 0.1875,0.8125, -0.1875,1.1875 },
 /* 22 */ { 0.068,0.684,0.726, 0.1922,0.8762, -0.0877,1.2183 },
 /* 23 */ { 0.147,0.735,0.661, 0.2058,0.9413, 0.0097,1.2354 },
 /* 24 */ { 0.233,0.778,0.583, 0.2276,1.0059, 0.1019,1.2393 },
 /* 25 */ { 0.325,0.812,0.484, 0.2563,1.0687, 0.1866,1.2312 },
 /* 26 */ { 0.419,0.839,0.348, 0.2904,1.1289, 0.2624,1.2127 },
 /* 27 */ { 0.514,0.857,0.000, 0.3285,1.1860, 0.3285,1.1860 },
 /* 28 */ { 0.000,0.562,0.827, 0.2188,0.7812, -0.2188,1.2188 },
 /* 29 */ { 0.069,0.621,0.781, 0.2239,0.8451, -0.1074,1.2592 },
 /* 30 */ { 0.149,0.671,0.726, 0.2390,0.9101, 0.0017,1.2830 },
 /* 31 */ { 0.237,0.712,0.661, 0.2628,0.9743, 0.1047,1.2906 },
 /* 32 */ { 0.330,0.742,0.583, 0.2938,1.0362, 0.1986,1.2837 },
 /* 33 */ { 0.425,0.765,0.484, 0.3300,1.0949, 0.2815,1.2649 },
 /* 34 */ { 0.520,0.780,0.348, 0.3700,1.1500, 0.3527,1.2367 },
 /* 35 */ { 0.614,0.789,0.000, 0.4123,1.2016, 0.4123,1.2016 },
 /* 36 */ { 0.000,0.500,0.866, 0.2500,0.7500, -0.2500,1.2500 },
 /* 37 */ { 0.070,0.558,0.827, 0.2558,0.8140, -0.1240,1.3023 },
 /* 38 */ { 0.152,0.606,0.781, 0.2726,0.8790, -0.0002,1.3337 },
 /* 39 */ { 0.241,0.644,0.726, 0.2988,0.9426, 0.1160,1.3449 },
 /* 40 */ { 0.335,0.671,0.661, 0.3323,1.0031, 0.2205,1.3385 },
 /* 41 */ { 0.431,0.689,0.583, 0.3708,1.0598, 0.3112,1.3182 },
 /* 42 */ { 0.525,0.700,0.484, 0.4125,1.1125, 0.3875,1.2875 },
 /* 43 */ { 0.617,0.706,0.348, 0.4559,1.1614, 0.4500,1.2496 },
 /* 44 */ { 0.707,0.707,0.000, 0.5000,1.2071, 0.5000,1.2071 },
 /* 45 */ { 0.000,0.438,0.899, 0.2812,0.7188, -0.2812,1.2812 },
 /* 46 */ { 0.071,0.495,0.866, 0.2879,0.7828, -0.1364,1.3485 },
 /* 47 */ { 0.155,0.541,0.827, 0.3068,0.8477, 0.0064,1.3886 },
 /* 48 */ { 0.246,0.574,0.781, 0.3359,0.9103, 0.1389,1.4027 },
 /* 49 */ { 0.341,0.597,0.726, 0.3721,0.9690, 0.2558,1.3954 },
 /* 50 */ { 0.436,0.610,0.661, 0.4128,1.0231, 0.3547,1.3719 },
 /* 51 */ { 0.529,0.617,0.583, 0.4559,1.0728, 0.4356,1.3372 },
 /* 52 */ { 0.619,0.619,0.484, 0.5000,1.1187, 0.5000,1.2955 },
 /* 53 */ { 0.706,0.617,0.348, 0.5441,1.1614, 0.5500,1.2496 },
 /* 54 */ { 0.789,0.614,0.000, 0.5877,1.2016, 0.5877,1.2016 },
 /* 55 */ { 0.000,0.375,0.927, 0.3125,0.6875, -0.3125,1.3125 },
 /* 56 */ { 0.072,0.432,0.899, 0.3202,0.7517, -0.1422,1.3991 },
 /* 57 */ { 0.158,0.474,0.866, 0.3419,0.8162, 0.0257,1.4487 },
 /* 58 */ { 0.252,0.503,0.827, 0.3742,0.8773, 0.1786,1.4643 },
 /* 59 */ { 0.347,0.520,0.781, 0.4133,0.9334, 0.3093,1.4534 },
 /* 60 */ { 0.440,0.528,0.726, 0.4560,0.9841, 0.4160,1.4243 },
 /* 61 */ { 0.530,0.530,0.661, 0.5000,1.0303, 0.5000,1.3839 },
 /* 62 */ { 0.617,0.529,0.583, 0.5441,1.0728, 0.5644,1.3372 },
 /* 63 */ { 0.700,0.525,0.484, 0.5875,1.1125, 0.6125,1.2875 },
 /* 64 */ { 0.780,0.520,0.348, 0.6300,1.1500, 0.6473,1.2367 },
 /* 65 */ { 0.857,0.514,0.000, 0.6715,1.1860, 0.6715,1.1860 },
 /* 66 */ { 0.000,0.312,0.950, 0.3438,0.6562, -0.3438,1.3438 },
 /* 67 */ { 0.074,0.368,0.927, 0.3529,0.7206, -0.1374,1.4561 },
 /* 68 */ { 0.162,0.406,0.899, 0.3781,0.7843, 0.0648,1.5155 },
 /* 69 */ { 0.257,0.429,0.866, 0.4143,0.8430, 0.2428,1.5290 },
 /* 70 */ { 0.351,0.439,0.827, 0.4561,0.8953, 0.3878,1.5102 },
 /* 71 */ { 0.442,0.442,0.781, 0.5000,0.9419, 0.5000,1.4723 },
 /* 72 */ { 0.528,0.440,0.726, 0.5440,0.9841, 0.5840,1.4243 },
 /* 73 */ { 0.610,0.436,0.661, 0.5872,1.0231, 0.6453,1.3719 },
 /* 74 */ { 0.689,0.431,0.583, 0.6292,1.0598, 0.6888,1.3182 },
 /* 75 */ { 0.765,0.425,0.484, 0.6700,1.0949, 0.7185,1.2649 },
 /* 76 */ { 0.839,0.419,0.348, 0.7096,1.1289, 0.7376,1.2127 },
 /* 77 */ { 0.910,0.414,0.000, 0.7483,1.1621, 0.7483,1.1621 },
 /* 78 */ { 0.000,0.250,0.968, 0.3750,0.6250, -0.3750,1.3750 },
 /* 79 */ { 0.076,0.303,0.950, 0.3863,0.6895, -0.1139,1.5232 },
 /* 80 */ { 0.168,0.335,0.927, 0.4161,0.7516, 0.1366,1.5901 },
 /* 81 */ { 0.263,0.350,0.899, 0.4563,0.8062, 0.3438,1.5938 },
 /* 82 */ { 0.354,0.354,0.866, 0.5000,0.8536, 0.5000,1.5607 },
 /* 83 */ { 0.439,0.351,0.827, 0.5439,0.8953, 0.6122,1.5102 },
 /* 84 */ { 0.520,0.347,0.781, 0.5867,0.9334, 0.6907,1.4534 },
 /* 85 */ { 0.597,0.341,0.726, 0.6279,0.9690, 0.7442,1.3954 },
 /* 86 */ { 0.671,0.335,0.661, 0.6677,1.0031, 0.7795,1.3385 },
 /* 87 */ { 0.742,0.330,0.583, 0.7062,1.0362, 0.8014,1.2837 },
 /* 88 */ { 0.812,0.325,0.484, 0.7437,1.0687, 0.8134,1.2312 },
 /* 89 */ { 0.881,0.320,0.348, 0.7803,1.1007, 0.8177,1.1808 },
 /* 90 */ { 0.949,0.316,0.000, 0.8162,1.1325, 0.8162,1.1325 },
 /* 91 */ { 0.000,0.188,0.982, 0.4062,0.5938, -0.4062,1.4062 },
 /* 92 */ { 0.079,0.237,0.968, 0.4209,0.6581, -0.0534,1.6068 },
 /* 93 */ { 0.173,0.260,0.950, 0.4567,0.7167, 0.2660,1.6701 },
 /* 94 */ { 0.265,0.265,0.927, 0.5000,0.7652, 0.5000,1.6490 },
 /* 95 */ { 0.350,0.263,0.899, 0.5437,0.8062, 0.6562,1.5938 },
 /* 96 */ { 0.429,0.257,0.866, 0.5857,0.8430, 0.7572,1.5290 },
 /* 97 */ { 0.503,0.252,0.827, 0.6258,0.8773, 0.8214,1.4643 },
 /* 98 */ { 0.574,0.246,0.781, 0.6641,0.9103, 0.8611,1.4027 },
 /* 99 */ { 0.644,0.241,0.726, 0.7012,0.9426, 0.8840,1.3449 },
 /* 100 */ { 0.712,0.237,0.661, 0.7372,0.9743, 0.8953,1.2906 },
 /* 101 */ { 0.778,0.233,0.583, 0.7724,1.0059, 0.8981,1.2393 },
 /* 102 */ { 0.844,0.230,0.484, 0.8070,1.0372, 0.8947,1.1907 },
 /* 103 */ { 0.910,0.227,0.348, 0.8411,1.0684, 0.8865,1.1442 },
 /* 104 */ { 0.974,0.225,0.000, 0.8748,1.0996, 0.8748,1.0996 },
 /* 105 */ { 0.000,0.125,0.992, 0.4375,0.5625, -0.4375,1.4375 },
 /* 106 */ { 0.084,0.168,0.982, 0.4581,0.6258, 0.0947,1.7159 },
 /* 107 */ { 0.177,0.177,0.968, 0.5000,0.6768, 0.5000,1.7374 },
 /* 108 */ { 0.260,0.173,0.950, 0.5433,0.7167, 0.7340,1.6701 },
 /* 109 */ { 0.335,0.168,0.927, 0.5839,0.7516, 0.8634,1.5901 },
 /* 110 */ { 0.406,0.162,0.899, 0.6219,0.7843, 0.9352,1.5155 },
 /* 111 */ { 0.474,0.158,0.866, 0.6581,0.8162, 0.9743,1.4487 },
 /* 112 */ { 0.541,0.155,0.827, 0.6932,0.8477, 0.9936,1.3886 },
 /* 113 */ { 0.606,0.152,0.781, 0.7274,0.8790, 1.0002,1.3337 },
 /* 114 */ { 0.671,0.149,0.726, 0.7610,0.9101, 0.9983,1.2830 },
 /* 115 */ { 0.735,0.147,0.661, 0.7942,0.9413, 0.9903,1.2354 },
 /* 116 */ { 0.799,0.145,0.583, 0.8270,0.9724, 0.9780,1.1904 },
 /* 117 */ { 0.863,0.144,0.484, 0.8596,1.0035, 0.9624,1.1473 },
 /* 118 */ { 0.927,0.143,0.348, 0.8920,1.0346, 0.9443,1.1059 },
 /* 119 */ { 0.990,0.141,0.000, 0.9243,1.0657, 0.9243,1.0657 },
 /* 120 */ { 0.000,0.062,0.998, 0.4688,0.5312, -0.4688,1.4688 },
 /* 121 */ { 0.088,0.088,0.992, 0.5000,0.5884, 0.5000,1.8258 },
 /* 122 */ { 0.168,0.084,0.982, 0.5419,0.6258, 0.9053,1.7159 },
 /* 123 */ { 0.237,0.079,0.968, 0.5791,0.6581, 1.0534,1.6068 },
 /* 124 */ { 0.303,0.076,0.950, 0.6137,0.6895, 1.1139,1.5232 },
 /* 125 */ { 0.368,0.074,0.927, 0.6471,0.7206, 1.1374,1.4561 },
 /* 126 */ { 0.432,0.072,0.899, 0.6798,0.7517, 1.1422,1.3991 },
 /* 127 */ { 0.495,0.071,0.866, 0.7121,0.7828, 1.1364,1.3485 },
 /* 128 */ { 0.558,0.070,0.827, 0.7442,0.8140, 1.1240,1.3023 },
 /* 129 */ { 0.621,0.069,0.781, 0.7761,0.8451, 1.1074,1.2592 },
 /* 130 */ { 0.684,0.068,0.726, 0.8078,0.8762, 1.0877,1.2183 },
 /* 131 */ { 0.747,0.068,0.661, 0.8395,0.9074, 1.0658,1.1790 },
 /* 132 */ { 0.810,0.067,0.583, 0.8711,0.9386, 1.0424,1.1410 },
 /* 133 */ { 0.872,0.067,0.484, 0.9027,0.9698, 1.0177,1.1040 },
 /* 134 */ { 0.935,0.067,0.348, 0.9342,1.0010, 0.9920,1.0677 },
 /* 135 */ { 0.998,0.067,0.000, 0.9656,1.0322, 0.9656,1.0322 },
 /* 136 */ { 0.000,0.000,1.000, 0.5000,0.5000, 0.5000,0.5000 },
 /* 137 */ { 0.062,0.000,0.998, 0.5312,0.5312, 1.4688,1.4688 },
 /* 138 */ { 0.125,0.000,0.992, 0.5625,0.5625, 1.4375,1.4375 },
 /* 139 */ { 0.188,0.000,0.982, 0.5938,0.5938, 1.4062,1.4062 },
 /* 140 */ { 0.250,0.000,0.968, 0.6250,0.6250, 1.3750,1.3750 },
 /* 141 */ { 0.312,0.000,0.950, 0.6562,0.6562, 1.3438,1.3438 },
 /* 142 */ { 0.375,0.000,0.927, 0.6875,0.6875, 1.3125,1.3125 },
 /* 143 */ { 0.438,0.000,0.899, 0.7188,0.7188, 1.2812,1.2812 },
 /* 144 */ { 0.500,0.000,0.866, 0.7500,0.7500, 1.2500,1.2500 },
 /* 145 */ { 0.562,0.000,0.827, 0.7812,0.7812, 1.2188,1.2188 },
 /* 146 */ { 0.625,0.000,0.781, 0.8125,0.8125, 1.1875,1.1875 },
 /* 147 */ { 0.688,0.000,0.726, 0.8438,0.8438, 1.1562,1.1562 },
 /* 148 */ { 0.750,0.000,0.661, 0.8750,0.8750, 1.1250,1.1250 },
 /* 149 */ { 0.812,0.000,0.583, 0.9062,0.9062, 1.0938,1.0938 },
 /* 150 */ { 0.875,0.000,0.484, 0.9375,0.9375, 1.0625,1.0625 },
 /* 151 */ { 0.938,0.000,0.348, 0.9688,0.9688, 1.0312,1.0312 },
 /* 152 */ { 1.000,0.000,0.000, 1.0000,1.0000, 1.0000,1.0000 }
};

static struct tilestrip_s
{
  int len, top, bottom;
}
tilestrips[] =
{
  { 1, 1, 0, },
  { 2, 3, 1, },
  { 3, 6, 3, },
  { 4, 10, 6, },
  { 5, 15, 10, },
  { 6, 21, 15, },
  { 7, 28, 21, },
  { 8, 36, 28, },
  { 9, 45, 36, },
  { 10, 55, 45, },
  { 11, 66, 55, },
  { 12, 78, 66, },
  { 13, 91, 78, },
  { 14, 105, 91, },
  { 15, 120, 105, },
  { 16, 136, 120, },
  { -1, -1, -1 }
};

static INLINE void DrawSkyTilePoint(const tilesky_info_t *info,
    const image_t *image, int p, int xmul, int ymul, int zmul, 
    float_t dx, float_t dy)
{
  float_t x = tilepoints[p].x * xmul;
  float_t y = tilepoints[p].y * ymul;
  float_t z = tilepoints[p].z * zmul;

  float_t tx = (zmul > 0) ? tilepoints[p].tx : tilepoints[p].bx;
  float_t ty = (zmul > 0) ? tilepoints[p].ty : tilepoints[p].by;

  if (xmul < 0)
  {
    float_t tmp = tx; tx = 1 - ty; ty = 1 - tmp;
  }
  
  if (ymul < 0)
  {
    float_t tmp = tx; tx = ty; ty = tmp;
  }
  
  CHECKVAL(info->number);
  CHECKVAL(info->squish);

  glTexCoord2f(tx * info->number + dx, ty * info->number + dy);
  glVertex3f(y * 1000, x * 1000, (z + info->offset) * 1000 / 
      info->squish);
}
    
static void DrawSkyTilePart(const tilesky_info_t *info,
    const image_t *image, int xmul, int ymul, int zmul, 
    float_t dx, float_t dy)
{
  int s;
  int len, top, bot;

  for (s=0; tilestrips[s].len > 0; s++)
  {
    glBegin(GL_TRIANGLE_STRIP);

    len = tilestrips[s].len;
    top = tilestrips[s].top;
    bot = tilestrips[s].bottom;

    for (; len > 0; len--, top++, bot++)
    {
      DrawSkyTilePoint(info, image, top, xmul, ymul, zmul, dx, dy);
      DrawSkyTilePoint(info, image, bot, xmul, ymul, zmul, dx, dy);
    }

    DrawSkyTilePoint(info, image, top, xmul, ymul, zmul, dx, dy);
    
    glEnd();
  }
}

static void RGL_DrawTiledSky(void)
{
  int i;
  const image_t *image;
  const cached_image_t *cim;
  const tilesky_info_t *info;

  side_t *side;
  float_t trans;
  float_t dx, dy;

  RGL_SetupMatricesTiledSky();

  for (i=0; i < MAX_TILESKY; i++)
  {
    if (! sky_tiles[i].active)
      continue;

    info = sky_tiles[i].info;
    side = sky_tiles[i].line->side[0];
    trans = side->middle.translucency;
    dx = side->middle.x_offset * info->number / 1024.0;
    dy = side->middle.y_offset * info->number / 1024.0;

    if (info->type == TILESKY_Flat)
      image = sky_tiles[i].line->frontsector->floor.image;
    else
      image = side->middle.image;

    if (!image || trans < 0.01)
      continue;

    glEnable(GL_TEXTURE_2D);

    cim = W_ImageCache(image, IMG_OGL, 0, true);
    glBindTexture(GL_TEXTURE_2D, W_ImageGetOGL(cim));

    if (trans <= 0.99 || !image->solid)
      glEnable(GL_BLEND);

    // sky is always 100% bright
    glColor4f(1.0, 1.0, 1.0, trans);

    DrawSkyTilePart(info, image,  1,  1,  1, dx, dy);
    DrawSkyTilePart(info, image, -1,  1,  1, dx, dy);
    DrawSkyTilePart(info, image, -1, -1,  1, dx, dy);
    DrawSkyTilePart(info, image,  1, -1,  1, dx, dy);

    DrawSkyTilePart(info, image,  1,  1, -1, dx, dy);
    DrawSkyTilePart(info, image, -1,  1, -1, dx, dy);
    DrawSkyTilePart(info, image, -1, -1, -1, dx, dy);
    DrawSkyTilePart(info, image,  1, -1, -1, dx, dy);

    W_ImageDone(cim);
    
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
  }
}

//
// RGL_DrawSky
//
void RGL_DrawSky(void)
{
  int x, y, w, h;
  float_t right, bottom;

  const cached_image_t *cim;
  GLuint tex_id;

  float_t mlook_rad;
  int top_L, bottom_L;
  int base_a = viewangle >> 6;

  int sx1, sy1, sx2, sy2;  // screen coords
  float_t tx1, tx2, ty, bx1, bx2, by;  // tex coords

  if (! draw_sky)
    return;

  // reset draw-sky flag
  draw_sky = false;
 
  if (sky_tiles_active > 0)
  {
    RGL_DrawTiledSky();
    return;
  }
  
  DEV_ASSERT2(sky_image);
  cim = W_ImageCache(sky_image, IMG_OGL, 0, true);
  tex_id = W_ImageGetOGL(cim);

  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, tex_id);

  w = sky_image->actual_w;
  h = sky_image->actual_h;
  right  = w / (float_t)sky_image->total_w;
  bottom = h / (float_t)sky_image->total_h;

  // sky is always 100% bright
  glColor4f(1.0, 1.0, 1.0, 1.0);

  sx1 = viewwindowx;
  sx2 = viewwindowx + viewwindowwidth;

  sy1 = viewwindowy;
  sy2 = viewwindowy + viewwindowheight;

  // compute sky horizontally tex coords
  mlook_rad = atan(viewvertangle);

  if (mlook_rad >= M_PI/4)
    top_L = ANG90;
  else if (mlook_rad <= -M_PI/4)
    top_L = ANG45;
  else
  {
    // d is just the distance horizontally forward from the eye to
    // the top edge of the view rectangle.
    float_t d = M_ROOT2 * sin(M_PI/2 - (mlook_rad + M_PI/4));

    top_L = ANG90 - M_ATan(d);
  }

  if (mlook_rad <= -M_PI/4)
    bottom_L = ANG90;
  else if (mlook_rad >= M_PI/4)
    bottom_L = ANG45;
  else
  {
    // d is just the distance horizontally forward from the eye to
    // the bottom edge of the view rectangle.
    float_t d = M_ROOT2 * sin(M_PI/2 - (M_PI/4 - mlook_rad));

    bottom_L = ANG90 - M_ATan(d);
  }
  
  #define NEWSKYSHIFT  ((float_t)(1 << (ANGLEBITS - 16)))

  top_L >>= 6;
  bottom_L >>= 6;

  CHECKVAL(w);
  tx1 = (float_t)(base_a + top_L) / NEWSKYSHIFT / (float_t)w;
  tx2 = (float_t)(base_a - top_L) / NEWSKYSHIFT / (float_t)w;

  bx1 = (float_t)(base_a + bottom_L) / NEWSKYSHIFT / (float_t)w;
  bx2 = (float_t)(base_a - bottom_L) / NEWSKYSHIFT / (float_t)w;

  if (w <= 512)
  {
    tx1 /= 2.0; tx2 /= 2.0;
    bx1 /= 2.0; bx2 /= 2.0;
  }

  // compute sky vertical tex coords
  {
    float_t top_a    = (M_PI/2 - mlook_rad - M_PI/4) / M_PI;
    float_t bottom_a = (M_PI/2 - mlook_rad + M_PI/4) / M_PI;
    
    if (top_a < 0)
      top_a = 0;

    if (bottom_a > 1.0)
      bottom_a = 1.0;
    
    DEV_ASSERT2(bottom_a > top_a);

    ty = top_a * bottom;
    by = bottom_a * bottom;
  }

  glBegin(GL_QUADS);
  
  // divide screen into many squares, to reduce distortion
  for (y=0; y < 8; y++)
  for (x=0; x < 8; x++)
  {
    int xa = sx1 + (sx2 - sx1) * x     / 8;
    int xb = sx1 + (sx2 - sx1) * (x+1) / 8;

    int ya = sy1 + (sy2 - sy1) * y     / 8;
    int yb = sy1 + (sy2 - sy1) * (y+1) / 8;
    
    float_t la = tx1 + (bx1 - tx1) * y     / 8;
    float_t ra = tx2 + (bx2 - tx2) * y     / 8;
    float_t lb = tx1 + (bx1 - tx1) * (y+1) / 8;
    float_t rb = tx2 + (bx2 - tx2) * (y+1) / 8;

    float_t txa = la + (ra - la) * x     / 8;
    float_t bxa = lb + (rb - lb) * x     / 8;
    float_t txb = la + (ra - la) * (x+1) / 8;
    float_t bxb = lb + (rb - lb) * (x+1) / 8;
    
    float_t tya = ty + (by - ty) * y     / 8;
    float_t bya = ty + (by - ty) * (y+1) / 8;
    
    glTexCoord2f(txa, 1.0 - bottom * tya);
    glVertex2i(xa, SCREENHEIGHT - ya);
    
    glTexCoord2f(txb, 1.0 - bottom * tya);
    glVertex2i(xb, SCREENHEIGHT - ya);
    
    glTexCoord2f(bxb, 1.0 - bottom * bya);
    glVertex2i(xb, SCREENHEIGHT - yb);
    
    glTexCoord2f(bxa, 1.0 - bottom * bya);
    glVertex2i(xa, SCREENHEIGHT - yb);
  }

  glEnd();
  
  glDisable(GL_TEXTURE_2D);

  W_ImageDone(cim);
}


#endif  // USE_GL
