//----------------------------------------------------------------------------
//  EDGE OpenGL Rendering (Textures)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------

// this conditional applies to the whole file
#ifdef USE_GL

#include "i_defs.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "e_search.h"
#include "m_bbox.h"
#include "m_random.h"
#include "p_local.h"
#include "p_mobj.h"
#include "r_defs.h"
#include "r_main.h"
#include "r_plane.h"
#include "r_sky.h"
#include "r_state.h"
#include "r_things.h"
#include "r2_defs.h"
#include "rgl_defs.h"
#include "v_colour.h"
#include "v_res.h"
#include "w_textur.h"
#include "w_wad.h"
#include "z_zone.h"


#define DEBUG_NO_TEX  0


//!!!!!! OLD CRUD
#define MAX_TEX  (10000)
stored_gl_tex_t conv_sprites[MAX_TEX];


static GLuint RGL_SendTexture(byte *gl_data, int width, int height)
{
  GLuint id;

  glEnable(GL_TEXTURE_2D);

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  glGenTextures(1, &id);

  glBindTexture(GL_TEXTURE_2D, id);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  glTexImage2D(GL_TEXTURE_2D, 0, 4, width, height,
               0, GL_RGBA, GL_UNSIGNED_BYTE, gl_data);

  glDisable(GL_TEXTURE_2D);

  return id;
}

// EXPERIMENTAL!
static byte *RGL_CreateHalo(int *width, int *height, float_t *right,
    float_t *bottom)
{
  byte *data;
  
  int x, y;
  
  *width = 64;
  *height = 64;
  *right = 1.0;
  *bottom = 1.0;

  data = Z_New(byte, 64*64*4);

  for (y=0; y < 64; y++)
  for (x=0; x < 64; x++)
  {
    byte *pix = &data[(y*64 + x) * 4];

    float_t dx = ABS(x - 32) / 32.0;
    float_t dy = ABS(y - 32) / 32.0;

    float_t dist = 1.0 - (dx + dy - MIN(dx,dy) / 2.0);

    int alpha = (int)(dist * dist * 255.0);
    
    pix[0] = pix[1] = pix[2] = 255;
    pix[3] = MIN(255, MAX(0, alpha));
  }

  return data;
}

//
// RGL_Sprite
//
void RGL_Sprite(int patch)
{
  int width, height;
  float_t right, bottom;

  byte *gl_data;

  if (patch == SPR_HALO)
    gl_data = RGL_CreateHalo(&width, &height, &right, &bottom);
  else
    gl_data = W_ConvertSpriteRGBA(patch, &width, &height, &right, &bottom);

  if (! gl_data)
  {
    conv_sprites[patch].status = TEX_DUMMY;
    conv_sprites[patch].id     = (M_Random() % 32) << 10;
    return;
  }

#if DEBUG_NO_TEX
  {
    conv_sprites[patch].status = TEX_DUMMY;
    conv_sprites[patch].id     = RGL_DummyColour(gl_data,
        floor(right * width), floor(bottom * height), width);

    Z_Free(gl_data);
    return;
  }
#endif

  conv_sprites[patch].status = TEX_OK;
  conv_sprites[patch].id     = RGL_SendTexture(gl_data, width, height);
  conv_sprites[patch].width  = width;
  conv_sprites[patch].height = height;
  conv_sprites[patch].right  = right;
  conv_sprites[patch].bottom = bottom;

  Z_Free(gl_data);
}


//
// RGL_InitTextures
//
void RGL_InitTextures(void)
{
  memset(conv_sprites,  0, sizeof(conv_sprites));
}

//
// RGL_ResetTextures
//
// Needed for gamma, since gamma is applied when textures are
// converted to RGBA format.

void RGL_ResetTextures(void)
{
  int i;
  GLuint tex_id;

  for (i=0; i < MAX_TEX; i++)
  {
    if (conv_sprites[i].status == TEX_OK)
    {
      tex_id = conv_sprites[i].id;
      glDeleteTextures(1, &tex_id);
    }
  }

  // clear texture caches
  RGL_InitTextures();
}


#endif  // USE_GL
