//----------------------------------------------------------------------------
//  EDGE New SaveGame Handling (Things)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// See the file "docs/save_sys.txt" for a complete description of the
// new savegame system.
//
// This file handles:
//    mobj_t        [MOBJ]
//    spawnspot_t   [SPWN]
//    shootspot_t   [SSPT]
//    iteminque_t   [ITMQ]
//
// TODO HERE:
//   + make `state' format more robust.
//

#include "i_defs.h"

#include "dm_state.h"
#include "e_player.h"
#include "p_local.h"
#include "sv_chunk.h"
#include "sv_main.h"
#include "r_state.h"
#include "z_zone.h"


// forward decls.
int SV_MobjCountElems(void);
int SV_MobjFindElem(mobj_t *elem);
void * SV_MobjGetElem(int index);
void SV_MobjCreateElems(int num_elems);
void SV_MobjFinaliseElems(void);

boolean_t SR_MobjGetPlayer(void *storage, int index, void *extra);
boolean_t SR_MobjGetMobj(void *storage, int index, void *extra);
boolean_t SR_MobjGetType(void *storage, int index, void *extra);
boolean_t SR_MobjGetState(void *storage, int index, void *extra);

void SR_MobjPutPlayer(void *storage, int index, void *extra);
void SR_MobjPutMobj(void *storage, int index, void *extra);
void SR_MobjPutType(void *storage, int index, void *extra);
void SR_MobjPutState(void *storage, int index, void *extra);


//
//  MOBJ STRUCTURE AND ARRAY
//
static mobj_t sv_dummy_mobj;

#define SV_F_BASE  sv_dummy_mobj

static savefield_t sv_fields_mobj[] =
{
  SVFIELD(x, "x", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(y, "y", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(z, "z", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(angle, "angle", 1, SVT_ANGLE, SR_GetAngle, SR_PutAngle),
  SVFIELD(floorz, "floorz", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(ceilingz, "ceilingz", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(dropoffz, "dropoffz", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(radius, "radius", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(height, "height", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(momx, "momx", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(momy, "momy", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(momz, "momz", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(health, "health", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(speed, "speed", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(fuse, "fuse", 1, SVT_INT, SR_GetInt, SR_PutInt),
  SVFIELD(info, "info", 1, SVT_STRING, SR_MobjGetType, SR_MobjPutType),
  SVFIELD(state, "state", 1, SVT_STRING, SR_MobjGetState, SR_MobjPutState),
  SVFIELD(next_state, "next_state", 1, SVT_STRING, SR_MobjGetState, SR_MobjPutState),
  SVFIELD(flags, "flags", 1, SVT_INT, SR_GetInt, SR_PutInt),
  SVFIELD(extendedflags, "extendedflags", 1, SVT_INT, SR_GetInt, SR_PutInt),
  SVFIELD(movedir, "movedir", 1, SVT_INT, SR_GetInt, SR_PutInt),
  SVFIELD(movecount, "movecount", 1, SVT_INT, SR_GetInt, SR_PutInt),
  SVFIELD(reactiontime, "reactiontime", 1, SVT_INT, SR_GetInt, SR_PutInt),
  SVFIELD(threshold, "threshold", 1, SVT_INT, SR_GetInt, SR_PutInt),
  SVFIELD(player, "player", 1, SVT_INDEX("players"), SR_MobjGetPlayer, SR_MobjPutPlayer),
// !!! SVFIELD(spawnpoint, "spawnpoint", 1, SVT_STRUCT("spawnpoint_t"), SR_MobjGetSpawnPoint, SR_MobjPutSpawnPoint),
  SVFIELD(origheight, "origheight", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(visibility, "visibility", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(vis_target, "vis_target", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(vertangle, "vertangle", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(spreadcount, "spreadcount", 1, SVT_INT, SR_GetInt, SR_PutInt),
  SVFIELD(side, "side", 1, SVT_INT, SR_GetInt, SR_PutInt),
///!!!  SVFIELD(currentattack, "currentattack", 1, SVT_FLOAT, SR_MobjGetAttack, SR_MobjPutAttack),
  SVFIELD(source, "source", 1, SVT_INDEX("mobjs"), SR_MobjGetMobj, SR_MobjPutMobj),
  SVFIELD(target, "target", 1, SVT_INDEX("mobjs"), SR_MobjGetMobj, SR_MobjPutMobj),
  SVFIELD(tracer, "tracer", 1, SVT_INDEX("mobjs"), SR_MobjGetMobj, SR_MobjPutMobj),
  SVFIELD(supportobj, "supportobj", 1, SVT_INDEX("mobjs"), SR_MobjGetMobj, SR_MobjPutMobj),
  SVFIELD(ride_em, "ride_em", 1, SVT_INDEX("mobjs"), SR_MobjGetMobj, SR_MobjPutMobj),
  SVFIELD(ride_dx, "ride_dx", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  SVFIELD(ride_dy, "ride_dy", 1, SVT_FLOAT, SR_GetFloat, SR_PutFloat),
  
// !!! path_trigger
// !!! shoot_info
 
  // NOT HERE:
  //   subsector & region: these are regenerated.
  //   next,prev,snext,sprev,bnext,bprev: links are regenerated.
  //   tunnel_hash: would be meaningless, and not important.
  //   lastlookup: being reset to zero won't hurt.
  //   ...
  
  SVFIELD_END
};

#undef SV_F_BASE

savestruct_t sv_struct_mobj =
{
  NULL,          // link in list
  "mobj_t",      // structure name
  "mobj",        // start marker
  sv_fields_mobj,   // field descriptions
  true,          // define_me
  NULL           // pointer to known struct
};

savearray_t sv_array_mobj =
{
  NULL,             // link in list
  "mobjs",          // array name
  &sv_struct_mobj,  // array type
  true,             // define_me

  SV_MobjCountElems,     // count routine
  SV_MobjGetElem,        // index routine
  SV_MobjCreateElems,    // creation routine
  SV_MobjFinaliseElems,  // finalisation routine

  NULL,     // pointer to known array
  0         // loaded size
};


//----------------------------------------------------------------------------

//
// SV_MobjCountElems
//
int SV_MobjCountElems(void)
{
  mobj_t *cur;
  int count=0;

  for (cur=mobjlisthead; cur; cur=cur->next)
    count++;
  
  return count;
}

//
// SV_MobjGetElem
//
void *SV_MobjGetElem(int index)
{
  mobj_t *cur;

  for (cur=mobjlisthead; cur && index > 0; cur=cur->next)
    index--;
  
  if (!cur)
    I_Error("LOADGAME: Invalid Mobj: %d\n", index);
  
  DEV_ASSERT2(index == 0);

  return cur;
}

//
// SV_MobjFindElem
//
int SV_MobjFindElem(mobj_t *elem)
{
  mobj_t *cur;
  int index;

  for (cur=mobjlisthead, index=0; cur && cur != elem; cur=cur->next)
    index++;
  
  if (!cur)
    I_Error("LOADGAME: No such MobjPtr: %p\n", elem);
  
  return index;
}

//
// SV_MobjCreateElems
//
void SV_MobjCreateElems(int num_elems)
{
  // free existing mobjs
  if (mobjlisthead)
    P_RemoveMobjs();
  
  mobjlisthead = NULL;
  
  for (; num_elems > 0; num_elems--)
  {
    mobj_t *cur = Z_ClearNew(mobj_t, 1);

    cur->next = mobjlisthead;
    cur->prev = NULL;

    if (mobjlisthead)
      mobjlisthead->prev = cur;
    
    mobjlisthead = cur;

    // initialise defaults
    cur->info = mobjinfo[0];
    cur->state = cur->next_state = states+1;
  }
}

//
// SV_MobjFinaliseElems
//
void SV_MobjFinaliseElems(void)
{
  mobj_t *mo;

  for (mo=mobjlisthead; mo; mo=mo->next)
  {
    if (mo->state)
    {
      mo->sprite = mo->state->sprite;
      mo->frame  = mo->state->frame;
      mo->bright = mo->state->bright;
    }

    P_SetThingPosition(mo);

    // handle reference counts

    #define REF_COUNT_FIELD(field)  \
        if (mo->field) mo->field->refcount++;
    
    REF_COUNT_FIELD(tracer);
    REF_COUNT_FIELD(source);
    REF_COUNT_FIELD(target);
    REF_COUNT_FIELD(supportobj);
    REF_COUNT_FIELD(ride_em);

    #undef REF_COUNT_FIELD

    // sanity checks
  }
}

//----------------------------------------------------------------------------

//
// SR_MobjGetPlayer
//
boolean_t SR_MobjGetPlayer(void *storage, int index, void *extra)
{
  player_t ** dest = (player_t **)storage + index;

  int swizzle = SV_GetInt();

  *dest = (swizzle == 0) ? NULL : SV_PlayerGetElem(swizzle - 1);
  return true;
}

//
// SR_MobjPutPlayer
//
void SR_MobjPutPlayer(void *storage, int index, void *extra)
{
  player_t *elem = ((player_t **)storage)[index];

  int swizzle = (elem == NULL) ? 0 : SV_PlayerFindElem(elem) + 1;

  SV_PutInt(swizzle);
}

//
// SR_MobjGetMobj
//
boolean_t SR_MobjGetMobj(void *storage, int index, void *extra)
{
  mobj_t ** dest = (mobj_t **)storage + index;

  int swizzle = SV_GetInt();

  *dest = (swizzle == 0) ? NULL : SV_MobjGetElem(swizzle - 1);
  return true;
}

//
// SR_MobjPutMobj
//
void SR_MobjPutMobj(void *storage, int index, void *extra)
{
  mobj_t *elem = ((mobj_t **)storage)[index];

  int swizzle;

  // -AJA- HACK: temp fix for dummy targets.
  if (elem && elem->extendedflags & EF_DUMMYMOBJ)
  {
    SV_PutInt(0);
    return;
  }

  swizzle = (elem == NULL) ? 0 : SV_MobjFindElem(elem) + 1;
  SV_PutInt(swizzle);
}

//
// SR_MobjGetType
//
boolean_t SR_MobjGetType(void *storage, int index, void *extra)
{
  mobjinfo_t ** dest = (mobjinfo_t **)storage + index;

  const char *name = SV_GetString();

  // Intentional Const Override
  *dest = (name == NULL) ? NULL : (mobjinfo_t *)DDF_MobjLookup(name);

  Z_Free((char *)name);
  return true;
}

//
// SR_MobjPutType
//
void SR_MobjPutType(void *storage, int index, void *extra)
{
  mobjinfo_t *info = ((mobjinfo_t **)storage)[index];

  SV_PutString(info->ddf.name);
}

//
// SR_MobjGetState
//
boolean_t SR_MobjGetState(void *storage, int index, void *extra)
{
  state_t ** dest = (state_t **)storage + index;

  const char *swizzle = SV_GetString();

  if (! swizzle)
  {
    *dest = NULL;
    return true;
  }

  //!!!! FIXME: make more robust
  *dest = states + atoi(swizzle);
  
  Z_Free((char *)swizzle);
  return true;
}

//
// SR_MobjPutState
//
void SR_MobjPutState(void *storage, int index, void *extra)
{
  state_t *S = ((state_t **)storage)[index];

  char swizzle[64];

  if (S == NULL)
  {
    SV_PutString(NULL);
    return;
  }

  //!!!! FIXME: make more robust
  sprintf(swizzle, "%d", S - states);
  SV_PutString(swizzle);
}

